// Code generated by "stringer -type=ExpandMode"; DO NOT EDIT.

package tflint

import "strconv"

func _() {
	// An "invalid array index" compiler error signifies that the constant values have changed.
	// Re-run the stringer command to generate them again.
	var x [1]struct{}
	_ = x[ExpandModeExpand-0]
	_ = x[ExpandModeNone-1]
}

const _ExpandMode_name = "ExpandModeExpandExpandModeNone"

var _ExpandMode_index = [...]uint8{0, 16, 30}

func (i ExpandMode) String() string {
	if i < 0 || i >= ExpandMode(len(_ExpandMode_index)-1) {
		return "ExpandMode(" + strconv.FormatInt(int64(i), 10) + ")"
	}
	return _ExpandMode_name[_ExpandMode_index[i]:_ExpandMode_index[i+1]]
}
