// SPDX-FileCopyrightText: Copyright 2015-2025 go-swagger maintainers
// SPDX-License-Identifier: Apache-2.0

package runtime

import "github.com/go-openapi/swag/fileutils"

type File = fileutils.File
