package plugin2host

import "github.com/terraform-linters/tflint-plugin-sdk/plugin/internal/plugin2host"

// Server is the interface that the host should implement when a plugin communicates with the host.
type Server = plugin2host.Server
